/****************************************************************************
 **
 ** Copyright (C) 2008-2009 Grigory A. Mozhaev.  All rights reserved.
 **
 ** This file may be used under the terms of the GNU General Public
 ** License version 2.0 as published by the Free Software Foundation
 ** and appearing in the file LICENSE.GPL included in the packaging of
 ** this file.  Please review the following information to ensure GNU
 ** General Public Licensing requirements will be met:
 ** http://www.trolltech.com/products/qt/opensource.html
 **
 ** If you are unsure which license is appropriate for your use, please
 ** review the following information:
 ** http://www.trolltech.com/products/qt/licensing.html or contact the
 ** sales department at sales@trolltech.com.
 **
 ** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 ** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 **
 ** AUTHOR: Grigory A. Mozhaev <zcrendel@mail.ru>
 **
 ****************************************************************************/
#define SOURCE_CODING "UTF-8"

#include <QtCore>
#include <QEvent>
#include <QtXml>

#include <CReport>
#include <CReportFactory>

uint CSection::mMaxId = 0;
QString CReport::mWorkDir = QString::null;

//! Static function
int CReport::generateReport(const QString &aScript, const QString &aWorkDir = QString::null) {
    mWorkDir = aWorkDir;

    if (!QSqlDatabase::drivers().contains("QSQLITE")) {
        qWarning("Unable to load database: needs the SQLITE driver");
        return 1;
    }

    QString vPath  = QDir::toNativeSeparators(QDir::currentPath()+"/tmp.db");

    if (QFile::exists(vPath)) {
        QFile::remove(vPath);
    }

    // initialize the database
    QSqlDatabase db; 
    if (QSqlDatabase::connectionNames().indexOf("qt_sql_default_connection") >= 0) {     
        db = QSqlDatabase::database();
    } else {
        db = QSqlDatabase::addDatabase("QSQLITE");
        db.setDatabaseName(vPath);
    }

    if (!db.open()) {
        qWarning("open failed: %s", db.lastError().text().toLocal8Bit().data());
        return 1;
    }

    QSqlQuery q;
    QString vQuery = QLatin1String("CREATE TABLE sections ("
                                   "id        integer,"
                                   "name      varchar,"
                                   "PRIMARY KEY(id))");
    if (!q.exec(vQuery)) {
        qDebug() << "stage 1:" << q.lastError().text();
        return 1;
    }

    vQuery = QLatin1String("CREATE TABLE rows ("
                           "section           integer,"
                           "row               integer,"
                           "tag               varchar,"
                           "value             varchar,"
                           "PRIMARY KEY (section, row, tag),"
                           "FOREIGN KEY (section) REFERENCES sections(id))");

    if (!q.exec(vQuery)) {
        qDebug() << "stage 2:" << q.lastError().text();
        return 1;
    }

    qWarning("sqlite database table was created!");
    qDebug() << vPath;

    QScriptEngine vEngine;
    addReportClasses(vEngine);

    QScriptValue vScriptValue = vEngine.evaluate(aScript);
    qDebug() << vEngine.uncaughtExceptionLineNumber() << ":" <<
        vScriptValue.toString();

    return 0;
}

//! Static function
void CReport::addReportClasses(QScriptEngine &aEngine) {
    QScriptValue vReportClass = aEngine.scriptValueFromQMetaObject<CReport>();
    aEngine.globalObject().setProperty("CReport", vReportClass);

    QScriptValue vSectionClass = aEngine.scriptValueFromQMetaObject<CSection>();
    aEngine.globalObject().setProperty("CSection", vSectionClass);

    QScriptValue vRowClass = aEngine.scriptValueFromQMetaObject<CRow>();
    aEngine.globalObject().setProperty("CRow", vRowClass);

    QScriptValue vTagClass = aEngine.scriptValueFromQMetaObject<CTag>();
    aEngine.globalObject().setProperty("CTag", vTagClass);

    qScriptRegisterMetaType<CTag*>(&aEngine,
                                    CTag::toScriptValue,
                                    CTag::fromScriptValue);
    qScriptRegisterMetaType<CRow*>(&aEngine,
                                    CRow::toScriptValue,
                                    CRow::fromScriptValue);
    qScriptRegisterMetaType<CSection*>(&aEngine,
                                    CSection::toScriptValue,
                                    CSection::fromScriptValue);
}

QString CReport::workDir() {
    return CReport::mWorkDir;
}


CReport::CReport(QObject *apParent) {
    qWarning("CReport()");

    mSections.setTable("sections");
    mRows.setTable("rows");

    mSections.setEditStrategy(QSqlTableModel::OnFieldChange);
    mRows.setEditStrategy(QSqlTableModel::OnFieldChange);
}

CReport::~CReport() {
    qWarning("~CReport()");
}

void CReport::setOption(const QString &aName, const QVariant &aValue) {
    qWarning("CReport::setOption() { ");

    qDebug() << "aName:" << aName;
    qDebug() << "aValue:" << aValue;

    qWarning("}");

    mOptions.insert(aName, aValue);
}

QString CReport::templatePath(const QString &aName) {
    return CReportFactory::templatePath(aName);
}

int CReport::generate(const QString &aReportType) {
    qWarning("CReport::generate() { ");

    qDebug() << "ReportType:" << aReportType;

    qWarning("}");

    IReport* vReport = CReportFactory::factory()->instance(aReportType);

    if (vReport != 0) {
        vReport->generate(this);
    } else {
        QMessageBox::critical(NULL, tr("Error"), 
                              tr("Plugin for report type '%1' not found'").arg(aReportType));
    }

    return 0;
}

CTag* CReport::tag(const QString &aName) {
    //qWarning("CReport::tag()");

    CSection *vSection = CReport::section();

    if (vSection) {
        CRow *vRow = vSection->row(1);
        if (vRow) {
            return vRow->tag(aName);
        }
    }

    return NULL;
}

CSection* CReport::section(const QString &aName) {
    //qWarning("CReport::section()");

    CSection *vSection = new CSection(this, aName);

    mSections.setFilter(QString("name='%1'").arg(aName));
    vSection->mId = 0;

    if ( mSections.select() ) {
        if (mSections.query().first()) {
            vSection->mId = mSections.record(0).value("id").toInt();
        }
    }

    return vSection;
}

// ********** CSection  ****************
CSection::CSection(QObject *apParent) 
    : QObject(apParent) {

}

CSection::CSection(CReport *apParent, const QString &aName) {
    //qWarning("CSection()");
    mName = aName;
    mId   = 0;
    mpReport = apParent;
}

CSection::~CSection() {
    //qWarning("~CSection()");
}

QString CSection::name() {    
    return mName;
}

CRow *CSection::row(uint aRow) {
    //qWarning("CSection::row()");

    CRow *vRow = new CRow(this, aRow);
    return vRow;
}

int CSection::rowCount() {
    //qWarning("CSection::rowCount(%d)", mId);
    if (mId > 0) {
        QSqlQueryModel model;
        model.setQuery(QString("SELECT max(row) FROM rows WHERE section=%1").arg(mId));
        return model.record(0).value(0).toUInt();
    }

    return 0;
}

QVariant CSection::min(const QString &aTagName) {
    qWarning("CSection::min()");
    return QVariant();
}

QVariant CSection::max(const QString &aTagName) {
    qWarning("CSection::max()");
    return QVariant();
}

QVariant CSection::avg(const QString &aTagName) {
    qWarning("CSection::avg()");
    return QVariant();
    
}

QVariant CSection::sum(const QString &aTagName) {
    qWarning("CSection::sum()");
    return QVariant();    
}

// **** CSection::private
QVariant CSection::value(uint aRow, const QString& aTagName) {
    //qWarning("CSection::value()");
    QVariant vTagValue;

    if (mId > 0) {
        mpReport->mRows.setFilter(QString("section=%1 AND row=%2 AND tag='%3'")
                        .arg(mId).arg(aRow).arg(aTagName));
        if ((mpReport->mRows.select()) && mpReport->mRows.query().first()) {
            vTagValue = mpReport->mRows.record(0).value("value").toString();
        } else {
            vTagValue = QVariant();
        }
    } else {
        vTagValue = QVariant();
    }

    return vTagValue;
}

void CSection::setValue(uint &aRow, const QString& aTagName,
                         const QVariant &aTagValue) {
    //qWarning("CSection::setValue()");
    if (mId == 0) {
        if (mName.isNull()) {
            mpReport->mSections.setFilter(QString("name is NULL"));
        } else {
            mpReport->mSections.setFilter(QString("name='%1'")
                                          .arg(mName));
        }
        bool vFlag = mpReport->mSections.select();
        //qDebug() << "Sections select : " << vFlag;
        if (vFlag) {
            if (mpReport->mSections.query().first()) {
                mId = mpReport->mSections.record(0).value("id").toInt();
            } else {
                mId = ++mMaxId;
            
                QSqlRecord vRecord;

                QSqlField  vFieldId("id", QVariant::UInt);
                vFieldId.setValue(mId);
                QSqlField  vFieldName("name", QVariant::String);

                if (!mName.isNull()) {                    
                    vFieldName.setValue(mName);
                } else {
                    vFieldName.setValue(QVariant(QVariant::String));
                }

                vRecord.append(vFieldId);
                vRecord.append(vFieldName);

                //qDebug() << vRecord;
        
                mpReport->mSections.insertRecord(-1, vRecord);

                if (!mpReport->mSections.submit())
                    qDebug() << mpReport->mSections.lastError().text();
            }
        } else {
            qDebug() << mpReport->mRows.lastError().text();
            return ;
        }
    }

    mpReport->mRows.setFilter(QString("section=%1 AND row=%2 AND tag='%3'")
                             .arg(mId).arg(aRow).arg(aTagName));

    bool vFlag = mpReport->mRows.select();
    //qDebug() << "Rows select : " << vFlag << mpReport->mRows.filter() << 
    //"|" << mpReport->mRows.query().lastQuery();
    if (vFlag) {
        //qWarning("Rows selected!");
        if (mpReport->mRows.query().first()) {
            mpReport->mRows.record(0).setValue("value", aTagValue);
            mpReport->mRows.submit();
            qDebug() << mpReport->mRows.lastError().text();
        } else {
            if (aRow == 0) {
                QSqlQueryModel model;
                model.setQuery(QString("SELECT max(row) FROM rows WHERE section='%1'").arg(mId));
                aRow = model.record(0).value("row").toUInt() + 1;
            }

            QSqlRecord vRecord;
            QSqlField  vFieldSec("section", QVariant::UInt);
            vFieldSec.setValue(mId);
        
            QSqlField  vFieldRow("row", QVariant::UInt);
            vFieldRow.setValue(aRow);

            QSqlField  vFieldName("tag", QVariant::String);
            vFieldName.setValue(aTagName);

            QSqlField  vFieldValue("value", QVariant::String);
            vFieldValue.setValue(aTagValue);

            vRecord.append(vFieldSec);
            vRecord.append(vFieldRow);
            vRecord.append(vFieldName);
            vRecord.append(vFieldValue);

            //qDebug() << vRecord;

            mpReport->mRows.insertRecord(-1, vRecord);

            if (!mpReport->mRows.submit())
                qDebug() << mpReport->mRows.lastError().text();
        }
    } else {
        qDebug() << mpReport->mRows.lastError().text();
    }
}

// *********************** CRow ******************************
CRow::CRow(QObject *apParent) 
    : QObject(apParent) {

}

CRow::CRow(CSection *apParent, uint aRow) {
    //qWarning("CRow()");
    mRow = aRow;
    mpSection = apParent;
}

CRow::~CRow() {
    //qWarning("~CRow()");
}

CTag *CRow::tag(const QString& aName) {
    CTag *vTag = new CTag(this, aName);
    return vTag;
}

void CRow::setValue(const QString& aTagName,
                     const QVariant &aTagValue) {
    //qWarning("CRow::setValue()");
    mpSection->setValue(mRow, aTagName, aTagValue);
}

QVariant CRow::value(const QString& aTagName) {
    //qWarning("CRow::value()");
    return mpSection->value(mRow, aTagName);
}

// ********************* CTag *******************************
CTag::CTag(QObject *apParent) 
    : QObject(apParent) {

}

CTag::CTag(CRow *apParent, const QString &aName) {
    //qWarning("CTag()");
    mpRow = apParent;
    mName = aName;
}

CTag::~CTag() {
    //qWarning("~CTag()");
}

QString CTag::name() {
    //qWarning("CTag::name()");
    return mName;
}

void CTag::setValue(const QVariant& aValue) {
    //qWarning("CTag::setValue()");
    mpRow->setValue(mName, aValue);
}

QVariant CTag::value() {
    //qWarning("CTag::value()");
    return mpRow->value(mName);
}

/*Tag& CTag::operator=(const QVariant& aValue) {
    //
}

Tag& CTag::operator=(const Tag& aTag) {
    //
}
*/
